"""
    partie.py
"""
import probleme
import string
import os
import random
import re

if "DEBUG" in os.environ:
    DEBUG = True
else:
    DEBUG = False

ALPHA = string.ascii_uppercase

class Partie:
    def __init__(self, dico):
        self.probleme = probleme.Probleme(dico)
        self.cle = "-" * len(ALPHA)
        self.set_hypothese()
        self.lettre_indice = ""
        self.dico = dico

    def reset(self):
        self.cle = "-" * len(ALPHA)
        self.set_hypothese()
        self.lettre_indice = ""
        
    def set_hypothese(self):
        ch = ""
        for car in self.probleme.crypto:
            if car.upper() in ALPHA:
                idx = ALPHA.find(car.upper())
                if self.cle[idx] != "-":
                    car = self.cle[idx]
                else:
                    car = "-"
            ch += car
        self.hypothese = ch

    def get_hypothese(self, largeur):
        chaine1 = []
        chaine2 = []
        position = 0
        msg = re.compile("\t").sub("    ",self.probleme.crypto) 
        hyp = re.compile("\t").sub("    ",self.hypothese ) 
        if msg[-1] in ALPHA:
            msg += " "
            hyp += " "
        lg = len(msg)
    
        while True:
            # -- on traite une tranche
            x = msg[position : position + largeur ].upper()
            last_pos = 0
            for i in range(len(x)):
                if x[i] == "\n":
                    last_pos = i
                    break
                if x[i] not in ALPHA:
                    last_pos = i
            if last_pos == 0 and i != 0:
                last_pos = len(x)
            xx = list(msg)
            #xx[ position + last_pos ] = "\n"
            msg2 = "".join(xx)
            yy = list(hyp)
            #yy[ position + last_pos ] = "\n"
            hyp2 = "".join(yy)
            chaine1.append( msg2[ position : position + last_pos + 1 ] )
            chaine2.append( hyp2[ position : position + last_pos + 1 ] )
            if (lg - (position + last_pos)) <= 1:
                break
            position += last_pos + 1
        chaine = ""
        for y in range(len(chaine1)):
            if chaine1[y][-1] != "\n":
                chaine1[y] += "\n"
                chaine2[y] += "\n"
            chaine += chaine1[y] + chaine2[y] + "\n"
        return chaine
        
    def get_hypo_pln(self):
        return chaine
    
    def get_hypo_cle(self):
        return ALPHA + "\n" + self.cle
        
    def new_hypothese(self, cry, pln):
        idx = ALPHA.find(cry)
        self.cle = self.cle[:idx] + pln.lower() + self.cle[idx+1:]
        self.set_hypothese()
        
    def est_ce_resolu(self):
        resolu = True
        for car in self.probleme.crypto:
            if car.upper() in ALPHA:
                idx = ALPHA.find(car.upper())
                if self.cle[idx] == "-":
                    resolu = False
                    break
        return resolu

    def hint(self):
        if not hasattr(self.probleme, 'cle'):
            return
        cle = self.probleme.cle
        if DEBUG:
            print("--- hint(): lettre_indice: ", self.lettre_indice)
        if len(self.lettre_indice) < 15:
            while True:
                alea = random.randint(0,25)
                if ALPHA[alea] not in self.lettre_indice:
                    break
            self.lettre_indice += ALPHA[alea]
            return cle[alea], ALPHA[alea]
        else:
            return

    def lg(self):
        lg = 0
        for car in (self.probleme.crypto).upper():
            if car in ALPHA:
                lg += 1
        return str(lg)
         
    def stats_unigram(self):
        freq = {}
        for car in (self.probleme.crypto).upper():
            if car in ALPHA:
                if car in freq:
                    freq[car] += 1
                else:
                    freq[car] = 1
        freq = dict(sorted(freq.items(), key=lambda item: item[1], reverse=True))
        ch = ""
        for x,y in freq.items():
            if type(y) == int:
                if y > 1:
                    ch += f"{x}->{y}, "
        return ch
        
    def stats_bigram(self):
        ch = ""
        for car in self.probleme.crypto:
            if car.upper() in ALPHA:
                ch += car
        dic = {}
        for i in range(0,len(ch)-1):
            digram = ch[i] + ch[i+1]
            if digram in dic:
                dic[digram] += 1
            else:
                dic[digram] = 1
        dic2 = list(dic.keys())
        for cle in dic2:
            if dic[cle] == 1:
                del dic[cle]
        dic2 = dict(sorted(dic.items(), key=lambda item: item[1], reverse=True))
        tb =list(dic2.items())[:20]
        ch = ""
        for couple in tb:
            ch += couple[0] + "=>" + str(couple[1]) + ", "
        return ch

    def stats_trigram(self):
        ch = ""
        for car in self.probleme.crypto:
            if car.upper() in ALPHA:
                ch += car
        dic = {}
        for i in range(0,len(ch)-2):
            trigram = ch[i] + ch[i+1] + ch[i+2]
            if trigram in dic:
                dic[trigram] += 1
            else:
                dic[trigram] = 1
        dic2 = list(dic.keys())
        for cle in dic2:
            if dic[cle] == 1:
                del dic[cle]
        dic2 = dict(sorted(dic.items(), key=lambda item: item[1], reverse=True))
        tb =list(dic2.items())[:20]
        ch = ""
        for couple in tb:
            ch += couple[0] + "=>" + str(couple[1]) + ", "
        return ch

    def get_dico(self):
        return self.dico
