# the_cry_gui.py , v1.4, JF 23/09/24

import tkinter as tk
import tkinter.messagebox as tkmsg
import tkinter.filedialog as tkfile
from tkinter.font import Font
import sys
import string
import json
import random
import pickle
import os
import re
import glob
import hashlib
import copy
import tarfile

import fenetres as fen
import polices
import partie

#--------------- fonctions Non GUI --------------------------
def affiche_occulte(dico):
    if dico["METHOD"] == "aristocrat" or dico["METHOD"] == "patristocrat":
        affiche_boutons()
    else:
        occulte_boutons()

def recharge():
    global PARTIE_COURANTE, DICO
    if os.access("./DATA/partie.dat", os.R_OK):
        PARTIE_COURANTE = pickle.load(open("./DATA/partie.dat","rb"))
        DICO = PARTIE_COURANTE.get_dico()
        if DEBUG:
            print("--- recharge(): Recharge partie en cours")   
            print("DICO: ", DICO)

def pro_alea():
    if not re.search(r'^..$', LANG):
        return None
    lst = glob.glob(f"./CRYPTOS/{LANG}/**/*.cry", recursive=True)
    if len(lst) == 0: return None
    return lst[ random.randint(0,len(lst)-1) ]

def sauve():
    if PARTIE_COURANTE:
        pickle.dump(PARTIE_COURANTE, open("./DATA/partie.dat","wb"))
        if DEBUG:
            print("-- sauve(): Partie sauvegardee")

def change_langue(x):
    global LANG
    LANG = x
    if DEBUG:
        print("--- change_langue(): Change langue courante: ", LANG)

def recharge_lang():
    global LST_LANG
    tb = glob.glob("./CRYPTOS/??")

    for fic in tb:
        m = re.search(r'\W(\w\w)$', fic)
        if m:
            LST_LANG.append( m.group(1) )
    if DEBUG:
        print("--- recharge_lang(): Liste des langues présentes dans CRYPTOS :", LST_LANG)    
        
#=============== globals ===========================
assert sys.version_info >= (3, 8)
if "DEBUG" in os.environ:
    DEBUG = True
else:
    DEBUG = False

ALPHA = string.ascii_uppercase
MINUSCULES = ALPHA.lower()+"-"
N = 26
POSITION = 0
OLD_POSITION = POSITION
POS_HYPO = 0
OLD_POS_HYPO = 0
X = None
Y = None
CRY = None
LARGEUR = 50
MODIF = False
SAISIE = ""
DICO = {}
DICO_NEW = {}
PARTIE_COURANTE = None

REP_COURANT = os.path.dirname(os.path.realpath(__file__))
if DEBUG:
    print("--- globals: Repertoire courant: ", REP_COURANT)
os.chdir(REP_COURANT)
try:
    if os.access("./DATA/lang.txt", os.R_OK):
        LANG = (open("./DATA/lang.txt").readline())[:2]
    else:
        LANG = 'EN'
except:
    LANG='EN'
if DEBUG:
    print("--- globals: LANG par défaut: ", LANG)
try:
    if os.access("./DATA/cmd_pdf.txt", os.R_OK):
        CMD_SHOW_PDF = (open("./DATA/cmd_pdf.txt").readline()).rstrip()
    else:
        CMD_SHOW_PDF = "exit "
except:
    CMD_SHOW_PDF = "exit "

recharge()
LST_LANG = []
recharge_lang()

#=============== fonctions GUI ==========================
def fenetre():
    global PARTIE_COURANTE, DICO
    inputDialog = fen.MyDialog(root)
    root.wait_window(inputDialog.top)
    if DEBUG:
        print('--- fenetre(): Texte saisie: ', str(DICO) )
    if   'PLAIN' in fen.DICO_NEW:
        SAISIE = fen.DICO_NEW['PLAIN']
    elif 'CRYPTO' in fen.DICO_NEW:
        SAISIE = fen.DICO_NEW['CRYPTO']
    else:
        SAISIE = ""
    if len(SAISIE) > 1:
        if ok_del_partie():
            aff_solution()
            DICO = copy.copy(fen.DICO_NEW)
            DICO["LANG"] = LANG
            PARTIE_COURANTE = partie.Partie(DICO)
            maj_gui()

def new_puzzle():
    """   Choix aleatoire d'un probleme  """
    global PARTIE_COURANTE, DICO
    chemin = pro_alea()
    if DEBUG:
        print("Problème : ", chemin)
    if chemin == None:
            tkmsg.showinfo("titre","There is no problem for the language "+LANG)
    else:
        if ok_del_partie():
            aff_solution()
            DICO = json.load(open(chemin))
            md5 = str(hashlib.md5((DICO["CRYPTO"]).encode('utf-8')).hexdigest())
            DICO["MD5"] = md5
            PARTIE_COURANTE = partie.Partie(DICO)
            if DEBUG:
                print("-- new_puzzle(): nouvelle Partie aleatoire --")
                print(PARTIE_COURANTE.get_dico())
            maj_gui()
            affiche_occulte(DICO)
            
def open_puzzle():
    global PARTIE_COURANTE, DICO
    fen_saisie = fen.FenetreSaisie(root, "MD5 value? ")
    root.wait_window(fen_saisie.top)
    SAISIE = fen.SAISIE
    if DEBUG:
        print('--- open_puzzle(): Texte saisie: ', SAISIE )
    if 'SAISIE' != "":
        lst = glob.glob(f"./CRYPTOS/??/**/{SAISIE}*.cry", recursive=True)
        lstbis = glob.glob(f"./CRYPTOS/WORK/{SAISIE}*.cry")
        lst += lstbis
        if DEBUG:
            print("===> Liste:", lst)
        if len(lst) == 0:
            tkmsg.showinfo("titre","No problem available")
            return
        if len(lst) == 1:
            if ok_del_partie():
                aff_solution()
                DICO = json.load(open(lst[0]))
                DICO["MD5"] = SAISIE
                PARTIE_COURANTE = partie.Partie(DICO)
                if DEBUG:
                    print("===> nouvelle Partie --")
                    print(DICO)
                maj_gui()
                affiche_occulte(DICO)
        else:
            tkmsg.showinfo("titre", "Ambiguous (MD5 too short)")
        
def maj_gui():
    texte.configure(state="normal")
    texte.delete("1.0", tk.END)
    if PARTIE_COURANTE:
        texte.insert("1.0", PARTIE_COURANTE.get_hypothese(LARGEUR))
        chaine = "\nKey: " + ALPHA + "\n" + "     " + PARTIE_COURANTE.cle
        lbl_key.configure(text = chaine)
    else:
        lbl_key.configure(text="\nKey: " + " "*26 + "\n")
    texte.configure(state="disabled")
    
def ok_del_partie():
    if PARTIE_COURANTE != None:
        reponse = tkmsg.askquestion("titre", "Do we delete the current game?")
        if reponse == "yes":
            return True
        else:
            return False
    else:
        return True
  
def restore():
    recharge()
    maj_gui()
                
def exit_crypt():
    sauve()
    root.destroy()
    if DEBUG:
        print("--- exit_crypt(): exit via le menu [pas par la croix]")

def reset():
    if PARTIE_COURANTE:
        PARTIE_COURANTE.reset()
        maj_gui()
        if DEBUG:
            print("--- reset(): DICO: ", DICO)
    
def crypt_close():
    global PARTIE_COURANTE
    if PARTIE_COURANTE:
        if not PARTIE_COURANTE.est_ce_resolu():
            reponse = tkmsg.askquestion("titre","Are you sure?(unresolved problem) ")
            if reponse != "yes":
                return
            else:
                if hasattr(PARTIE_COURANTE.probleme, "plain"):
                    tkmsg.showinfo("Solution:", \
                        PARTIE_COURANTE.probleme.plain)
        PARTIE_COURANTE = None
        if os.access("./DATA/partie.dat", os.R_OK):
            os.remove("./DATA/partie.dat")
            if DEBUG:
                print("--- crypt_close(): Partie remanente detruite") 
        maj_gui()

def change_index(x):
    global POSITION, OLD_POSITION, POS_HYPO, OLD_POS_HYPO
    if POSITION == x: return
    OLD_POSITION = POSITION
    POSITION = x
    OLD_POS_HYPO = POS_HYPO
    POS_HYPO = OLD_POS_HYPO
    maj_alpha()         
    maj_hypo()
    
def maj_alpha():
    LST_ALPHA[OLD_POSITION].configure(bg=DEFAULT_BG)
    LST_ALPHA[POSITION].configure(bg='yellow')

def hypothese(x):
    global POS_HYPO, OLD_POS_HYPO
    OLD_POS_HYPO = POS_HYPO
    POS_HYPO = x
    maj_hypo()
    if PARTIE_COURANTE:
        PARTIE_COURANTE.new_hypothese(ALPHA[POSITION], MINUSCULES[x])
        maj_hypo()
    maj_gui()

def maj_hypo():
    LST_HYPO[POS_HYPO].configure(bg='red')
    LST_HYPO[OLD_POS_HYPO].configure(bg=DEFAULT_BG)

def aff_stats():
    if PARTIE_COURANTE:
        ch = ""
        ch += "Length: " + PARTIE_COURANTE.lg() + "\n\n"
        ch += "Unigrams: " + PARTIE_COURANTE.stats_unigram() + "\n\n"
        ch += "Bigrams: " + PARTIE_COURANTE.stats_bigram() + "\n\n"
        ch += "Trigrams: " + PARTIE_COURANTE.stats_trigram() 
        tkmsg.showinfo("Stats", ch)
        
def aff_hint():
    if PARTIE_COURANTE:
        couple = PARTIE_COURANTE.hint()
        if DEBUG:
            print("--- aff_hint(): Hint: ", couple)
        if couple != None:
            tkmsg.showinfo("Hint", couple[0] + "->" + couple[1].lower() )

def aff_begin():
    if PARTIE_COURANTE:
        if not hasattr(PARTIE_COURANTE.probleme, 'plain'):
            return
        ch = ""
        i = 0
        for car in PARTIE_COURANTE.probleme.plain:
            if car.upper() in ALPHA:
                ch += car
                i +=1 
                if i > 2:
                    break
            else:
                ch += car
        if i == 3:
            tkmsg.showinfo("Begin", "Beginning: " + ch )

def plain_notfound():
    if PARTIE_COURANTE:
        if not hasattr(PARTIE_COURANTE.probleme, 'plain'):
            return
        ch = ""
        for car in ALPHA.lower():
            if car not in PARTIE_COURANTE.cle:
                ch += car
        if len(ch) != 0:
            tkmsg.showinfo("Key", "Plain letters not found: " + ch )

def aff_meta():
    if PARTIE_COURANTE:
        ch = ""
        for cle,val in DICO.items():
            if cle in ["CRYPTO", "PLAIN", "KEY"]:
                continue
            else:
                ch += cle + " => " + str(val) + "\n"
        tkmsg.showinfo("Metadata", ch)
            
def is_solved():
    if PARTIE_COURANTE:
        if not hasattr(PARTIE_COURANTE.probleme, 'plain'):
            return
        solved = True
        for car in PARTIE_COURANTE.probleme.crypto:
            if car.upper() in ALPHA:
                idx = ALPHA.find(car.upper())
                if PARTIE_COURANTE.cle[idx] == "-":
                    solved = False
                    break
        tkmsg.showinfo("Is Solved?", "Yes" if solved else "No")

def is_key_correct():
    if PARTIE_COURANTE:
        if not hasattr(PARTIE_COURANTE.probleme, 'cle'):
            return
        key_correct = True
        la_cle = PARTIE_COURANTE.probleme.cle
        reverse_cle = [""]*len(ALPHA)
        for i in range(26):
            car = la_cle[i]
            j = ALPHA.find(car.upper())
            reverse_cle[j] = ALPHA[i]
        reverse_cle = "".join(reverse_cle)
        for i in range(len(ALPHA)):
            car = PARTIE_COURANTE.cle[i] 
            if car != "-":
                if car.upper() != reverse_cle[i].upper():
                    key_correct = False
        tkmsg.showinfo("Is Key correct?", "Yes" if key_correct else "No")
           
def aff_solution():
    if PARTIE_COURANTE:
        if hasattr(PARTIE_COURANTE.probleme, 'plain'):
            ch = PARTIE_COURANTE.probleme.plain
            tkmsg.showinfo("Solution", ch )

def to_clip():
    if PARTIE_COURANTE:
        lg = PARTIE_COURANTE.lg()
        unigrams = PARTIE_COURANTE.stats_unigram()
        bigrams  = PARTIE_COURANTE.stats_bigram()
        trigrams = PARTIE_COURANTE.stats_trigram()
        crypto = PARTIE_COURANTE.probleme.crypto
        ch = ""
        
        msg = "MD5: " + \
            str(hashlib.md5((crypto).encode('utf-8')).hexdigest())
        msg += "\n\nCryptogram:\n\n"
        msg += PARTIE_COURANTE.get_hypothese(LARGEUR)
        msg += "\n\n"
        if DICO["METHOD"] in ["aristocrat", "patristocrat"]:
            msg += "Key:" + "\n" + ALPHA + "\n" + PARTIE_COURANTE.cle + "\n\n"
        msg += "Length: " + str(lg) + "\n\n\n"
        msg += "Statistics:\n" 
        msg += "-Unigrams: " + unigrams + "\n\n"
        msg += "-Bigrams: " + bigrams + "\n\n"
        msg += "-Trigrams: " + trigrams + "\n\n"
        msg += "\n\nMeta-Data:\n"
        for cle, valeur in DICO.items():
            if cle in ["KEY", "KEYWORD", "PLAIN", "CRYPTO", "HINT_PLAIN", "MD5"]:
                continue
            msg += cle + ": " + str(valeur) + "\n"
        root.clipboard_clear()
        root.clipboard_append(msg)
        
def aff_pdf(type_fic):
    fichier = "./DATA/" + LANG + type_fic + ".pdf"
    existe = False
    if os.access(fichier, os.R_OK):
        existe = True
    else:
        fichier = "./DATA/EN" + type_fic + ".pdf"
        if os.access(fichier, os.R_OK):
            existe = True
    if existe:
        if sys.platform == 'linux':
            final = " &"
        else:
            final = ""
        cmd = CMD_SHOW_PDF + " "  + fichier + final
        os.system(cmd)

def lst_challenges():
    global PARTIE_COURANTE, DICO
    try:
        inputDialog = fen.LstChall(root)
        root.wait_window(inputDialog.top)
    except AttributeError:
        if DEBUG:
            print("AttributeError Exception")
    else:
        if len(fen.DICO_NEW) == 0:
            return
        if ok_del_partie():
            aff_solution()
            DICO = fen.DICO_NEW
            PARTIE_COURANTE = partie.Partie(DICO)
            if DEBUG:
                print("--- lst_challenges(): nouvelle Partie")
                print(DICO)
            maj_gui()
            affiche_occulte(DICO)

def load_file():
    chemin = tkfile.askopenfilename()
    if DEBUG:
        print("--- load_file: File:", chemin)
    if chemin == "":
        return
    if not re.search("^TC.*\.tar", os.path.basename(chemin), re.I):
        tkmsg.showinfo("Warning", "This File isn't a TheCry file")
        return
    tar = tarfile.open(chemin)
    tar.extract("HEADER.json", path="./TMP")
    f = open("./TMP/HEADER.json", "r")
    dico = json.load(f)
    f.close()
    if DEBUG:
        print("--- load_file(): File Type: ", dico["TYPE"])
    if dico["TYPE"] == "CHALLENGE":
        if os.listdir('./CRYPTOS/CHALL'):
            reponse = tkmsg.askquestion("titre", "Do you want to delete present Challenge?")
            if reponse == "yes":
                folder = './CRYPTOS/CHALL/'
                for filename in os.listdir(folder):
                    file_path = os.path.join(folder, filename)
                    os.remove(file_path)
            else:
                return        
        tar.extractall(path="./CRYPTOS/CHALL/")
        tkmsg.showinfo("Info", "A new Challenge was loaded")
    elif dico["TYPE"] == "LANG":
        lst_fic = []
        for tarinfo in tar:
            if not re.search("\.cry$", tarinfo.name, re.I):
                continue
            lst_fic.append( tarinfo.name )
            tar.extract(tarinfo, path="./CRYPTOS/")
        f = open("./CRYPTOS/FILES_ADDED/" + dico["LANG"] + "_" + dico["TITLE"], "w")
        f.write("\n".join(lst_fic))
        f.close()
        if dico["LANG"] not in LST_LANG:
            tkmsg.showinfo("Warning", "Restart the program to add the language")
        else:
            tkmsg.showinfo("Warning", "New files (for a language) were loaded")
    else:
        tkmsg.showinfo("Warning", "The file format is incorrect")
        return
   
def delete_file():
    try:
        inputDialog = fen.FileToDelete(root)
        root.wait_window(inputDialog.top)
    except AttributeError:
        if DEBUG:
            print("--- delete_file: AttributeError Exception")
    else:
        pass
        
def affiche_boutons():
    frame_alpha.pack()
    frame_letter.pack()   
    frame_key.pack()

def occulte_boutons():
    frame_alpha.pack_forget()
    frame_letter.pack_forget()      
    frame_key.pack_forget()  

def text_about():
    the_text = "TheCry V1.4\n"
    the_text += "Author: JF Bouchaudy\n"
    the_text += "License: GPL\n"
    the_text += 'Home Page: http://www.jfbouch.fr/crypto/thecry'
    tkmsg.showinfo("Info", the_text)
    
#========== interface graphique =====================
#------fenetre racine 
root = tk.Tk()
root.title("CRYPT")

#----- creation d'un menu
def fake():
    pass
m = tk.Menu(root)
sm1 = tk.Menu(root)
sm2 = tk.Menu(root)
sm3 = tk.Menu(root)
sm4 = tk.Menu(root)
sm5 = tk.Menu(root)
sm6 = tk.Menu(root)
m.add_cascade(label="File", menu=sm1)
m.add_cascade(label="Information", menu=sm2)
m.add_cascade(label="Human", menu=sm3)
m.add_cascade(label="Computer", menu=sm4)
m.add_cascade(label="Challenge", menu=sm5)
m.add_cascade(label="Help", menu=sm6)

sm1.add_command(label="New random Puzzle", command=new_puzzle)
sm1.add_command(label="Open Puzzle via MD5", command=open_puzzle)
sm1.add_separator()
sm1.add_command(label="Reset", command=reset)
sm1.add_command(label="Save", command=sauve)
sm1.add_command(label="Restore", command=restore)
sm1.add_command(label="Close", command=crypt_close)
sm1.add_separator()
sm1.add_command(label="To Clipboard", command=to_clip)
sm1.add_separator()
sm1.add_command(label="Exit", command=exit_crypt)

sm2.add_command(label="Print Statistics", command=aff_stats)
sm2.add_command(label="Hint", command=aff_hint)
sm2.add_command(label="Beginning", command=aff_begin)
sm2.add_command(label="Meta-Data", command=aff_meta)
sm2.add_command(label="Plain letters not found", command=plain_notfound)
sm2.add_separator()
sm2.add_command(label="Is sovled?", command=is_solved)
sm2.add_command(label="Is Key Correct?", command=is_key_correct)
sm2.add_command(label="Solution", command=aff_solution)

sm3.add_command(label="New Puzzle", command=fenetre)

sm4.add_command(label="New Puzzle", command=new_puzzle)
sm4.add_separator()
i = 0 ; x = tk.IntVar()
for langue in LST_LANG:
    sm4.add_radiobutton(label=langue, value=i, var = x, \
        command=lambda x=langue: change_langue(x))
    if langue == LANG:
        x.set(i)
    i += 1
sm5.add_command(label="List challenges", command=lst_challenges)
sm5.add_separator()
sm5.add_command(label="Load file", command=load_file)
sm5.add_command(label="Delete files", command=delete_file)
sm6.add_command(label="Manual", \
    command = lambda x="_manual": aff_pdf(x))
sm6.add_separator()
sm6.add_command(label="Statistics", \
    command = lambda x="_stats": aff_pdf(x))
sm6.add_separator()
sm6.add_command(label="About", command = text_about)
root.config(menu=m)

#----- creation d'une police
ma_police = polices.cree_police()
DEFAULT_BG = root.cget('bg')

#----- fenetre frame cryptogram
frame_crypto = tk.Frame(root)
texte = tk.Text(frame_crypto, width = 50, heigh = 9, font=ma_police)
scrol = tk.Scrollbar(frame_crypto, orient="vertical")
scrol.config(command=texte.yview)
texte.configure(yscrollcommand = scrol.set )
texte.pack(side='left')
scrol.pack(side='right', fill='y')
frame_crypto.pack()

#----- fenetre frame alphabets
frame_alpha = tk.Frame(root)
frame_letter = tk.Frame(root)
frame_key = tk.Frame(root)

LST_ALPHA = []
tk.Label(frame_alpha, text='CIPHER  ', font=ma_police).pack(side='left')
for i in range(N):
   w = tk.Button(frame_alpha, text=ALPHA[i], font=Font(family='Courier',size='11'),
        command=lambda x=i: change_index(x))
   w.pack(side='left')
   LST_ALPHA.append(w)
tk.Button(frame_alpha, text=" ", font=Font(family='Courier',size='11')).pack(side='left')

LST_HYPO = []
tk.Label(frame_letter, text='plain   ', font=ma_police).pack(side='left')
for i in range(N+1):
    w = tk.Button(frame_letter, text=MINUSCULES[i], 
       font=Font(family='Courier',size='11'), 
       command=lambda x=i: hypothese(x))
    w.pack(side='left')
    LST_HYPO.append(w)

lbl_key = tk.Label(frame_key, text="Key: ", font=ma_police)   
lbl_key.pack()

affiche_boutons()
if "METHOD" in DICO:
    affiche_occulte(DICO)
maj_gui()

#-------> on rentre dans boucle evenements
root.mainloop()
